@@ -39,6 +39,8 @@ static void rhythmdb_query_model_tree_model_init (GtkTreeModelIface *iface);
 static void rhythmdb_query_model_drag_source_init (RbTreeDragSourceIface *iface);
 static void rhythmdb_query_model_drag_dest_init (RbTreeDragDestIface *iface);
 static void rhythmdb_query_model_init (RhythmDBQueryModel *shell_player);
+static GObject *rhythmdb_query_model_constructor (GType type, guint n_construct_properties,
+						  GObjectConstructParam *construct_properties);
 static void rhythmdb_query_model_finalize (GObject *object);
 static void rhythmdb_query_model_set_property (GObject *object,
 					       guint prop_id,
@@ -75,6 +77,11 @@ static gboolean rhythmdb_query_model_row_drop_possible (RbTreeDragDest *drag_des
 							GtkTreePath *dest,
 							GtkTreeViewDropPosition pos,
 							GtkSelectionData  *selection_data);
+static gboolean rhythmdb_query_model_row_drop_position (RbTreeDragDest   *drag_dest,
+							GtkTreePath       *dest_path,
+							GList *targets,
+							GtkTreeViewDropPosition *pos);
+
 static GtkTreeModelFlags rhythmdb_query_model_get_flags (GtkTreeModel *model);
 static gint rhythmdb_query_model_get_n_columns (GtkTreeModel *tree_model);
 static GType rhythmdb_query_model_get_column_type (GtkTreeModel *tree_model, int index);
@@ -110,12 +117,13 @@ static GtkTargetList *rhythmdb_query_model_drag_target_list = NULL;
 struct RhythmDBQueryModelUpdate
 {
 	enum {
-		RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED,
+		RHYTHMDB_QUERY_MODEL_UPDATE_ROWS_INSERTED,
 		RHYTHMDB_QUERY_MODEL_UPDATE_ROW_CHANGED,
 		RHYTHMDB_QUERY_MODEL_UPDATE_ROW_DELETED,
 		RHYTHMDB_QUERY_MODEL_UPDATE_QUERY_COMPLETE,
 	} type;
 	RhythmDBEntry *entry;
+	GPtrArray *entries;
 
 	/* Only used for _ROW_CHANGED */
 	RhythmDBPropType prop;
@@ -254,6 +262,7 @@ rhythmdb_query_model_class_init (RhythmDBQueryModelClass *klass)
 	object_class->get_property = rhythmdb_query_model_get_property;
 
 	object_class->finalize = rhythmdb_query_model_finalize;
+	object_class->constructor = rhythmdb_query_model_constructor;
 
 	g_object_class_install_property (object_class,
 					 PROP_RHYTHMDB,
@@ -349,7 +358,7 @@ rhythmdb_query_model_drag_dest_init (RbTreeDragDestIface *iface)
 {
 	iface->drag_data_received = rhythmdb_query_model_drag_data_received;
 	iface->row_drop_possible = rhythmdb_query_model_row_drop_possible;
-	iface->row_drop_position = NULL;
+	iface->row_drop_position = rhythmdb_query_model_row_drop_position;
 }
 
 static void
@@ -365,22 +374,6 @@ rhythmdb_query_model_set_property (GObject *object,
 	case PROP_RHYTHMDB:
 	{
 		model->priv->db = g_value_get_object (value);
-		g_signal_connect_object (G_OBJECT (model->priv->db),
-					 "entry_added",
-					 G_CALLBACK (rhythmdb_query_model_entry_added_cb),
-					 model, 0);
-		g_signal_connect_object (G_OBJECT (model->priv->db),
-					 "entry_restored",
-					 G_CALLBACK (rhythmdb_query_model_entry_added_cb),
-					 model, 0);
-		g_signal_connect_object (G_OBJECT (model->priv->db),
-					 "entry_changed",
-					 G_CALLBACK (rhythmdb_query_model_entry_changed_cb),
-					 model, 0);
-		g_signal_connect_object (G_OBJECT (model->priv->db),
-					 "entry_deleted",
-					 G_CALLBACK (rhythmdb_query_model_entry_deleted_cb),
-					 model, 0);
 		break;
 	}
 	case PROP_QUERY:
@@ -454,6 +447,40 @@ rhythmdb_query_model_init (RhythmDBQueryModel *model)
 	model->priv->reorder_drag_and_drop = FALSE;
 
 	model->priv->model_poll_id = g_idle_add ((GSourceFunc) idle_poll_model, model);
+
+}
+
+static GObject *
+rhythmdb_query_model_constructor (GType type, guint n_construct_properties,
+				  GObjectConstructParam *construct_properties)
+{
+	RhythmDBQueryModel *model;
+	RhythmDBQueryModelClass *klass;
+	GObjectClass *parent_class;  
+
+	klass = RHYTHMDB_QUERY_MODEL_CLASS (g_type_class_peek (RHYTHMDB_TYPE_QUERY_MODEL));
+
+	parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
+	model = RHYTHMDB_QUERY_MODEL (parent_class->constructor (type, n_construct_properties,
+								 construct_properties));
+	g_signal_connect_object (G_OBJECT (model->priv->db),
+				 "entry_added",
+				 G_CALLBACK (rhythmdb_query_model_entry_added_cb),
+				 model, 0);
+	g_signal_connect_object (G_OBJECT (model->priv->db),
+				 "entry_restored",
+				 G_CALLBACK (rhythmdb_query_model_entry_added_cb),
+				 model, 0);
+	g_signal_connect_object (G_OBJECT (model->priv->db),
+				 "entry_changed",
+				 G_CALLBACK (rhythmdb_query_model_entry_changed_cb),
+				 model, 0);
+	g_signal_connect_object (G_OBJECT (model->priv->db),
+				 "entry_deleted",
+				 G_CALLBACK (rhythmdb_query_model_entry_deleted_cb),
+				 model, 0);
+
+	return G_OBJECT (model);
 }
 
 static void
@@ -461,9 +488,15 @@ rhythmdb_query_model_free_pending_update (RhythmDBQueryModel *model,
 					  struct RhythmDBQueryModelUpdate *update)
 {
 	switch (update->type) {
-	case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED:
-		rhythmdb_entry_unref (model->priv->db, update->entry);
+	case RHYTHMDB_QUERY_MODEL_UPDATE_ROWS_INSERTED:
+	{
+		guint i;
+		for (i = 0; i < update->entries->len; i++)
+			rhythmdb_entry_unref (model->priv->db,
+					      g_ptr_array_index (update->entries, i));
 		break;
+		g_ptr_array_free (update->entries, TRUE);
+	}
 	case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_CHANGED:
 		rhythmdb_entry_unref (model->priv->db, update->entry);
 		g_value_unset (&update->old);
@@ -608,6 +641,12 @@ rhythmdb_query_model_entry_changed_cb (RhythmDB *db, RhythmDBEntry *entry,
 	if (g_hash_table_lookup (model->priv->reverse_map, entry) != NULL) {
 		struct RhythmDBQueryModelUpdate *update;
 
+		if (model->priv->query &&
+		    !rhythmdb_evaluate_query (db, model->priv->query, entry)) {
+			rhythmdb_query_model_remove_entry (model, entry);
+			return;
+		}
+
 		rb_debug ("queueing entry change");
 
 		update = g_new0 (struct RhythmDBQueryModelUpdate, 1);
@@ -639,9 +678,10 @@ rhythmdb_query_model_entry_deleted_cb (RhythmDB *db, RhythmDBEntry *entry,
 /* Threading: Called from the database context, holding a db write lock
  */
 void
-rhythmdb_query_model_add_entry (RhythmDBQueryModel *model, RhythmDBEntry *entry)
+rhythmdb_query_model_add_entries (RhythmDBQueryModel *model, GPtrArray *entries)
 {
 	struct RhythmDBQueryModelUpdate *update;
+	guint i;
 
 	if (model->priv->max_count > 0
 	    && g_hash_table_size (model->priv->reverse_map) >= model->priv->max_count)
@@ -649,15 +689,26 @@ rhythmdb_query_model_add_entry (RhythmDBQueryModel *model, RhythmDBEntry *entry)
 	/* Check size later */
 
 	update = g_new (struct RhythmDBQueryModelUpdate, 1);
-	update->type = RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED;
-	update->entry = entry;
+	update->type = RHYTHMDB_QUERY_MODEL_UPDATE_ROWS_INSERTED;
+	update->entries = entries;
 
 	/* Called with a locked database */
-	rhythmdb_entry_ref_unlocked (model->priv->db, entry);
+	for (i = 0; i < update->entries->len; i++)
+		rhythmdb_entry_ref_unlocked (model->priv->db,
+					     g_ptr_array_index (update->entries, i));
 
 	g_async_queue_push (model->priv->pending_updates, update);
 }
 
+void
+rhythmdb_query_model_add_entry (RhythmDBQueryModel *model, RhythmDBEntry *entry)
+{
+	GPtrArray *entries = g_ptr_array_new ();
+
+	g_ptr_array_add (entries, entry);
+	rhythmdb_query_model_add_entries (model, entries);
+}
+
 
 /* Threading: Called from the database context, holding a db write lock
  */
@@ -843,9 +894,12 @@ rhythmdb_query_model_poll (RhythmDBQueryModel *model, GTimeVal *timeout)
 		iter.stamp = model->priv->stamp;
 
 		switch (update->type) {
-		case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_INSERTED:
+		case RHYTHMDB_QUERY_MODEL_UPDATE_ROWS_INSERTED:
 		{
-			rhythmdb_query_model_do_insert (model, update->entry);
+			guint i;
+			rb_debug ("inserting %d rows", update->entries->len);
+			for (i = 0; i < update->entries->len; i++)
+				rhythmdb_query_model_do_insert (model, g_ptr_array_index (update->entries, i));
 			break;
 		}
 		case RHYTHMDB_QUERY_MODEL_UPDATE_ROW_CHANGED:
@@ -1122,6 +1176,15 @@ rhythmdb_query_model_row_drop_possible (RbTreeDragDest *drag_dest,
 	return model->priv->sort_func == NULL;
 }
 
+static gboolean
+rhythmdb_query_model_row_drop_position (RbTreeDragDest   *drag_dest,
+					GtkTreePath       *dest_path,
+					GList *targets,
+					GtkTreeViewDropPosition *pos)
+{
+	return TRUE;
+}
+
 static GtkTreeModelFlags
 rhythmdb_query_model_get_flags (GtkTreeModel *model)
 {
