@@ -73,6 +73,7 @@ static void rb_playlist_manager_cmd_edit_automatic_playlist (BonoboUIComponent *
 							     const char *verbname);
 static void handle_playlist_entry_into_playlist_cb (RBPlaylist *playlist, const char *uri, const char *title,
 						    const char *genre, RBPlaylistManager *mgr);
+static gboolean reap_dead_playlist_threads (RBPlaylistManager *mgr);
 
 struct RBPlaylistManagerPrivate
 {
@@ -88,9 +89,22 @@ struct RBPlaylistManagerPrivate
 
 	GList *playlists;
 
+	guint playlist_serial;
+
 	RBPlaylistSource *loading_playlist;
 
 	char *firsturi;
+
+	guint thread_reaper_id;
+
+	GAsyncQueue *status_queue;
+	gint outstanding_threads;
+
+	GCond *saving_condition;
+	GMutex *saving_mutex;
+
+	gboolean saving;
+	gboolean dirty;
 };
 
 enum
@@ -258,8 +272,38 @@ static void
 rb_playlist_manager_init (RBPlaylistManager *mgr)
 {
 	mgr->priv = g_new0 (RBPlaylistManagerPrivate, 1);
+
+	mgr->priv->status_queue = g_async_queue_new ();
+
+	mgr->priv->saving_condition = g_cond_new ();
+	mgr->priv->saving_mutex = g_mutex_new ();
+
+	mgr->priv->saving = FALSE;
+	mgr->priv->dirty = FALSE;
+
+	mgr->priv->thread_reaper_id = g_idle_add ((GSourceFunc) reap_dead_playlist_threads, mgr);
 }
 
+static gboolean
+reap_dead_playlist_threads (RBPlaylistManager *mgr)
+{
+	GObject *obj;
+
+	while ((obj = g_async_queue_try_pop (mgr->priv->status_queue)) != NULL) {
+		GDK_THREADS_ENTER ();
+		g_object_unref (obj);
+		mgr->priv->outstanding_threads--;
+		GDK_THREADS_LEAVE ();
+	}
+
+	GDK_THREADS_ENTER ();
+	mgr->priv->thread_reaper_id = g_timeout_add (5000, (GSourceFunc) reap_dead_playlist_threads, mgr);
+	GDK_THREADS_LEAVE ();
+	
+	return FALSE;
+}
+
+
 static void
 rb_playlist_manager_finalize (GObject *object)
 {
@@ -272,6 +316,18 @@ rb_playlist_manager_finalize (GObject *object)
 
 	g_return_if_fail (mgr->priv != NULL);
 
+	g_source_remove (mgr->priv->thread_reaper_id);
+	
+	while (mgr->priv->outstanding_threads > 0) {
+		GObject *obj = g_async_queue_pop (mgr->priv->status_queue);
+		g_object_unref (obj);
+		mgr->priv->outstanding_threads--;
+	}
+	
+	g_async_queue_unref (mgr->priv->status_queue);
+
+	g_mutex_free (mgr->priv->saving_mutex);
+	g_cond_free (mgr->priv->saving_condition);
 
 	g_list_free (mgr->priv->playlists);
 
@@ -412,9 +468,9 @@ rb_playlist_manager_parse_file (RBPlaylistManager *mgr, const char *uri)
 	{
 		RBPlaylist *parser = rb_playlist_new ();
 
-		g_signal_connect (G_OBJECT (parser), "entry",
-				  G_CALLBACK (handle_playlist_entry_into_playlist_cb),
-				  mgr);
+		g_signal_connect_object (G_OBJECT (parser), "entry",
+					 G_CALLBACK (handle_playlist_entry_into_playlist_cb),
+					 mgr, 0);
 
 		if (!rb_playlist_parse (parser, uri))
 			rb_error_dialog (_("Couldn't parse playlist"));
@@ -438,8 +494,8 @@ static void
 append_new_playlist_source (RBPlaylistManager *mgr, RBPlaylistSource *source)
 {
 	mgr->priv->playlists = g_list_append (mgr->priv->playlists, source);
-	g_signal_connect (G_OBJECT (source), "deleted",
-			  G_CALLBACK (rb_playlist_manager_source_deleted_cb), mgr);
+	g_signal_connect_object (G_OBJECT (source), "deleted",
+				 G_CALLBACK (rb_playlist_manager_source_deleted_cb), mgr, 0);
 	g_signal_emit (G_OBJECT (mgr), rb_playlist_manager_signals[PLAYLIST_ADDED], 0,
 		       source);
 }
@@ -478,30 +534,88 @@ out:
 	g_free (file);
 }
 
+struct RBPlaylistManagerSaveThreadData
+{
+	RBPlaylistManager *mgr;
+	xmlDocPtr doc;
+};
+
+static gpointer
+rb_playlist_manager_save_thread_main (struct RBPlaylistManagerSaveThreadData *data)
+{
+	char *file;
+	char *tmpname;
+
+	rb_debug ("entering save thread");
+
+	file = g_build_filename (rb_dot_dir (), "playlists.xml", NULL);
+	tmpname = g_strconcat (file, ".tmp", NULL);
+	
+	xmlSaveFormatFile (tmpname, data->doc, 1);
+	rename (tmpname, file);
+	xmlFreeDoc (data->doc);
+	g_free (tmpname);
+	g_free (file);
+
+	g_mutex_lock (data->mgr->priv->saving_mutex);
+
+	data->mgr->priv->saving = FALSE;
+
+	g_cond_signal (data->mgr->priv->saving_condition);
+	g_mutex_unlock (data->mgr->priv->saving_mutex);
+
+	g_async_queue_push (data->mgr->priv->status_queue, data->mgr);
+
+	g_free (data);
+	
+	return NULL;
+}
+
 void
 rb_playlist_manager_save_playlists (RBPlaylistManager *mgr)
 {
-	char *file = g_build_filename (rb_dot_dir (), "playlists.xml", NULL);
-	char *tmpname = g_strconcat (file, ".tmp", NULL);
 	GList *tmp;
-	xmlDocPtr doc;
 	xmlNodePtr root;
+	struct RBPlaylistManagerSaveThreadData *data;
 
-	rb_debug ("saving playlists");
+	rb_debug ("saving the playlists");
 
-	doc = xmlNewDoc ("1.0");
+	if (!mgr->priv->dirty) {
+		rb_debug ("no save needed, ignoring");
+		return;
+	}
 
-	root = xmlNewDocNode (doc, NULL, "rhythmdb-playlists", NULL);
-	xmlDocSetRootElement (doc, root);
+	g_mutex_lock (mgr->priv->saving_mutex);
+
+	while (mgr->priv->saving)
+		g_cond_wait (mgr->priv->saving_condition, mgr->priv->saving_mutex);
+
+	mgr->priv->saving = TRUE;
+
+	g_mutex_unlock (mgr->priv->saving_mutex);
+
+	data = g_new0 (struct RBPlaylistManagerSaveThreadData, 1);
+	data->mgr = mgr;
+	data->doc = xmlNewDoc ("1.0");
+
+	root = xmlNewDocNode (data->doc, NULL, "rhythmdb-playlists", NULL);
+	xmlDocSetRootElement (data->doc, root);
 	
 	for (tmp = mgr->priv->playlists; tmp; tmp = tmp->next)
 		rb_playlist_source_save_to_xml (RB_PLAYLIST_SOURCE (tmp->data), root);
 
-	xmlSaveFormatFile (tmpname, doc, 1);
-	rename (tmpname, file);
-	xmlFreeDoc (doc);
-	g_free (tmpname);
-	g_free (file);
+	g_object_ref (G_OBJECT (mgr));
+	mgr->priv->outstanding_threads++;
+	
+	g_thread_create ((GThreadFunc) rb_playlist_manager_save_thread_main, data, FALSE, NULL);
+}
+
+static void
+rb_playlist_manager_set_dirty (RBPlaylistManager *mgr)
+{
+	g_mutex_lock (mgr->priv->saving_mutex);
+	mgr->priv->dirty = TRUE;
+	g_mutex_unlock (mgr->priv->saving_mutex);
 }
 
 RBSource *
@@ -509,9 +623,23 @@ rb_playlist_manager_new_playlist (RBPlaylistManager *mgr,
 				  const char *suggested_name, gboolean automatic)
 {
 	RBSource *playlist = RB_SOURCE (rb_playlist_source_new (mgr->priv->db, automatic));
-	g_object_set (G_OBJECT (playlist), "name", suggested_name ? suggested_name : "", NULL);
+	GTimeVal serial;
+	char *internal;
+
+
+	g_get_current_time (&serial);
+	internal = g_strdup_printf ("<playlist:%ld:%ld>", serial.tv_sec,
+				    serial.tv_usec);
+	
+	g_object_set (G_OBJECT (playlist),
+		      "name", suggested_name ? suggested_name : "",
+		      "internal-name", internal,
+		      NULL);
+	g_free (internal);
 	append_new_playlist_source (mgr, RB_PLAYLIST_SOURCE (playlist));
 	rb_sourcelist_edit_source_name (mgr->priv->sourcelist, playlist);
+	rb_playlist_manager_set_dirty (mgr);
+
 	return playlist;
 }
 
@@ -556,10 +684,14 @@ rb_playlist_manager_cmd_new_automatic_playlist (BonoboUIComponent *component,
 		gtk_widget_destroy (GTK_WIDGET (creator));	
 		return;
 	}
+
+
 	playlist = rb_playlist_manager_new_playlist (mgr, NULL, TRUE);
 
 	rb_playlist_manager_set_automatic_playlist (mgr, RB_PLAYLIST_SOURCE (playlist), creator); 
 
+	rb_playlist_manager_set_dirty (mgr);
+
 	gtk_widget_destroy (GTK_WIDGET (creator));	
 }
 
@@ -596,6 +728,7 @@ rb_playlist_manager_cmd_delete_playlist (BonoboUIComponent *component,
 	rb_debug ("Deleting playlist %p", mgr->priv->selected_source);
 	
 	rb_source_delete_thyself (mgr->priv->selected_source);
+	rb_playlist_manager_set_dirty (mgr);
 }
 
 static void
@@ -603,25 +736,32 @@ load_playlist_response_cb (GtkDialog *dialog,
 			   int response_id,
 			   RBPlaylistManager *mgr)
 {
-	char *file, *escaped_file;
+	char *escaped_file = NULL;
 
 	if (response_id != GTK_RESPONSE_OK) {
 		gtk_widget_destroy (GTK_WIDGET (dialog));
 		return;
 	}
 
+#ifndef HAVE_GTK_2_3
+	{
+	char *file = NULL;
 	file = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION (dialog)));
+	escaped_file = gnome_vfs_get_uri_from_local_path (file);
+	g_free (file);
+	}
+#else
+	escaped_file = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));
+#endif
 
 	gtk_widget_destroy (GTK_WIDGET (dialog));
 
-	if (file == NULL)
+	if (escaped_file == NULL)
 		return;
 
-	escaped_file = gnome_vfs_get_uri_from_local_path (file);
-	g_free (file);
-
 	rb_playlist_manager_parse_file (mgr, escaped_file);
 	g_free (escaped_file);
+	rb_playlist_manager_set_dirty (mgr);
 }
 
 static void
@@ -631,11 +771,11 @@ rb_playlist_manager_cmd_load_playlist (BonoboUIComponent *component,
 {
 	GtkWidget *dialog;
 
-	dialog = rb_ask_file_multiple (_("Load playlist"), NULL,
-			              GTK_WINDOW (mgr->priv->window));
+	dialog = rb_ask_file (_("Load playlist"), NULL,
+			      GTK_WINDOW (mgr->priv->window));
 
-	g_signal_connect (G_OBJECT (dialog), "response",
-			  G_CALLBACK (load_playlist_response_cb), mgr);
+	g_signal_connect_object (G_OBJECT (dialog), "response",
+				 G_CALLBACK (load_playlist_response_cb), mgr, 0);
 }
 
 static void
@@ -668,12 +808,12 @@ rb_playlist_manager_cmd_save_playlist (BonoboUIComponent *component,
 {
 	GtkWidget *dialog;
     
-	dialog = rb_ask_file_multiple (_("Save playlist"), NULL,
-			              GTK_WINDOW (mgr->priv->window));
+	dialog = rb_ask_file_save (_("Save playlist"), NULL,
+				   GTK_WINDOW (mgr->priv->window));
 
-	g_signal_connect (G_OBJECT (dialog), "response",
-			  G_CALLBACK (save_playlist_response_cb),
-			  mgr);
+	g_signal_connect_object (G_OBJECT (dialog), "response",
+				 G_CALLBACK (save_playlist_response_cb),
+				 mgr, 0);
 }
 
 static void
@@ -700,14 +840,34 @@ out:
 }
 
 static void
-handle_playlist_entry_into_playlist_cb (RBPlaylist *playlist, const char *uri, const char *title,
+handle_playlist_entry_into_playlist_cb (RBPlaylist *playlist, const char *uri_maybe,
+					const char *title,
 					const char *genre, RBPlaylistManager *mgr)
 {
+	char *uri;
+
+	if (uri_maybe[0] == '/') {
+		uri = gnome_vfs_get_uri_from_local_path (uri_maybe);
+		if (uri == NULL) {
+			rb_debug ("Error processing absolute filename: %s", uri_maybe);
+			return;
+		}
+	} else {
+		GnomeVFSURI *vfsuri = gnome_vfs_uri_new (uri_maybe);
+		if (!vfsuri) {
+			rb_debug ("Error processing probable URI: %s", uri_maybe);
+			return;
+		}
+		uri = gnome_vfs_uri_to_string (vfsuri, GNOME_VFS_URI_HIDE_NONE);
+		gnome_vfs_uri_unref (vfsuri);
+	}
+	
 	if (!mgr->priv->firsturi)
 		mgr->priv->firsturi = g_strdup (uri);
 	if (rb_uri_is_iradio (uri)) {
 		rb_iradio_source_add_station (mgr->priv->iradio_source,
 					      uri, title, genre);
+		g_free (uri);
 		return;
 	}
 
@@ -716,4 +876,5 @@ handle_playlist_entry_into_playlist_cb (RBPlaylist *playlist, const char *uri, c
 			RB_PLAYLIST_SOURCE (rb_playlist_manager_new_playlist (mgr, NULL, FALSE));
 	}
 	add_uri_to_playlist (mgr, mgr->priv->loading_playlist, uri, title);
+	g_free (uri);
 }
